/***************************************************************************
 *   Copyright (C) 2007-2010 by Tobias Schwarz                             *
 *   mail@tobias-schwarz.net                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   In addition, as a special exception, the copyright holders give       *
 *   permission to link the code of this program with any edition of       *
 *   the Qt library by Trolltech AS, Norway (or with modified versions     *
 *   of Qt that use the same license as Qt), and distribute linked         *
 *   combinations including the two.  You must obey the GNU General        *
 *   Public License in all respects for all of the code used other than    *
 *   Qt.  If you modify this file, you may extend this exception to        *
 *   your version of the file, but you are not obligated to do so.  If     *
 *   you do not wish to do so, delete this exception statement from        *
 *   your version.                                                         *
 ***************************************************************************/

#include "programWindowClass.h"
#include "kemboss.h"
#include "main.h"
#include "configDlg.h"
#include "associated_qualifiers.h"
#include <cstdio>

kprogramWindow::kprogramWindow(kembossWindow *parent) : kembossWindow(parent)
{
   setAttribute(Qt::WA_DeleteOnClose);
   kembossParent = parent;
   kpwInputTextEdit = NULL;
   kpwInputFileLineEdit = NULL;
   kpwOutputFileLineEdit = NULL;
   kpwOutputTextEdit = NULL;
   programProgress = NULL;
   highlightInput = 0;
   highlightOutput = 0;
   highlighter = new Highlighter();

   // Standard output as well as standard error data are both shown in the programProgress window
   programProcess.setProcessChannelMode(QProcess::MergedChannels);

   // The program progress window is updated whenever new data is available
   connect(&programProcess, SIGNAL(readyReadStandardOutput()), this, SLOT(updateProgramProgressForm()));
   connect(&programProcess, SIGNAL(finished(int,QProcess::ExitStatus)), this, SLOT(processFinished(int)));
   // connect(&loadThread, SIGNAL(loadedFile(QString, bool)), this, SLOT(appendText(QString, bool)));

   loadPrinterSettings();
}

void kprogramWindow::postWindowCreated()
{
   createActions();
   createMenus();
   createToolbars();
   findControls();
   readSettings();
   tabChanged(0);
   setTextEditFont();
   createRecentFiles();
   setEmbossApp(objectName().remove(0,1)); // The EMBOSS application could be changed in beforeRunningProgram() to allow single GUI windows to start different programs
   if (qobject_cast<kemboss *>(kembossParent)->ksettings.openLastSettings()) {
      QSettings settings("kemboss.org", objectName());
      QString fileName = settings.value("last_settings").toString();
      if (QFile::exists(fileName) && loadFile(fileName)) {
         setCurrentFile(fileName);
         return;
      }
   }
   newSettings();
}

void kprogramWindow::createActions()
{
   newAct = new QAction(QIcon(":/new/prefix1/filenew.png"), tr("&New"), this);
   newAct->setShortcut(tr("Ctrl+N"));
   newAct->setStatusTip(tr("Set to default settings"));
   connect(newAct, SIGNAL(triggered()), this, SLOT(newSettings()));

   openAct = new QAction(QIcon(":/new/prefix1/fileopen.png"), tr("&Open..."), this);
   openAct->setShortcut(tr("Ctrl+O"));
   openAct->setStatusTip(tr("Open settings from file"));
   connect(openAct, SIGNAL(triggered()), this, SLOT(open()));

   saveAct = new QAction(QIcon(":/new/prefix1/filesave.png"), tr("&Save"), this);
   saveAct->setShortcut(tr("Ctrl+S"));
   saveAct->setStatusTip(tr("Save settings to current file"));
   connect(saveAct, SIGNAL(triggered()), this, SLOT(save()));
   saveAct->setEnabled(false);

   saveAsAct = new QAction(QIcon(":/new/prefix1/filesaveas.png"), tr("Save &As..."), this);
   saveAsAct->setStatusTip(tr("Save settings to new file"));
   connect(saveAsAct, SIGNAL(triggered()), this, SLOT(saveAs()));

   exitAct = new QAction(QIcon(":/new/prefix1/exit.png"), tr("&Quit"), this);
   exitAct->setShortcut(tr("Ctrl+Q"));
   exitAct->setStatusTip(tr("Close window"));
   connect(exitAct, SIGNAL(triggered()), this, SLOT(close()));

   propertiesAct = new QAction(QIcon(":/new/prefix1/document-properties.png"), tr("Print properties"), this);
   propertiesAct->setStatusTip(tr("Print properties"));
   connect(propertiesAct, SIGNAL(triggered()), this, SLOT(documentProperties()));

   printAct = new QAction(QIcon(":/new/prefix1/document-print.png"), tr("Print"), this);
   printAct->setShortcut(tr("Ctrl+P"));
   printAct->setStatusTip(tr("Print"));
   connect(printAct, SIGNAL(triggered()), this, SLOT(printDocument()));

   runAct = new QAction(QIcon(":/new/prefix1/run.png"), tr("&Run"), this);
   runAct->setShortcut(tr("Ctrl+R"));
   runAct->setStatusTip(tr("Execute EMBOSS application"));
   connect(runAct, SIGNAL(triggered()), this, SLOT(runEmbossApp()));

   addParameterAct = new QAction(QIcon(":/new/prefix1/add.png"), tr("Add Parameter"), this);
   addParameterAct->setStatusTip(tr("Add configuration parameter for EMBOSS application"));
   connect(addParameterAct, SIGNAL(triggered()), this, SLOT(addParameter()));

   copyCommandlineAct = new QAction(QIcon(":/new/prefix1/document-export.png"), tr("Export commandline"), this);
   copyCommandlineAct->setStatusTip(tr("Export commandline for EMBOSS application to clipboard"));
   connect(copyCommandlineAct, SIGNAL(triggered()), this, SLOT(copyCommandline()));

   cutAct = new QAction(QIcon(":/new/prefix1/cut.png"), tr("Cu&t"), this);
   cutAct->setShortcut(tr("Ctrl+X"));
   cutAct->setStatusTip(tr("Cut the current selection's contents to the clipboard"));
   cutAct->setEnabled(false);

   copyAct = new QAction(QIcon(":/new/prefix1/copy.png"), tr("&Copy"), this);
   copyAct->setShortcut(tr("Ctrl+C"));
   copyAct->setStatusTip(tr("Copy the current selection's contents to the clipboard"));
   copyAct->setEnabled(false);

   pasteAct = new QAction(QIcon(":/new/prefix1/paste.png"), tr("&Paste"), this);
   pasteAct->setShortcut(tr("Ctrl+V"));
   pasteAct->setStatusTip(tr("Paste the clipboard's contents into the current selection"));

   undoAct = new QAction(QIcon(":/new/prefix1/undo.png"), tr("&Undo"), this);
   undoAct->setShortcut(tr("Ctrl+Z"));
   undoAct->setStatusTip(tr("Undo last operation"));
   undoAct->setEnabled(false);

   redoAct = new QAction(QIcon(":/new/prefix1/redo.png"), tr("&Redo"), this);
   QList<QKeySequence> redoShortcuts;
   redoShortcuts << tr("Ctrl+Y") << tr("Shift+Ctrl+Z");
   redoAct->setShortcuts(redoShortcuts);
   redoAct->setStatusTip(tr("Redo last operation"));
   redoAct->setEnabled(false);

   rotateLeftAct = new QAction(QIcon(":/new/prefix1/object-rotate-left.png"), tr("Rotate left"), this);
   rotateLeftAct->setStatusTip(tr("Rotate graph left"));
   connect(rotateLeftAct, SIGNAL(triggered()), this, SLOT(rotateLeft()));

   rotateRightAct = new QAction(QIcon(":/new/prefix1/object-rotate-right.png"), tr("Rotate right"), this);
   rotateRightAct->setStatusTip(tr("Rotate graph right"));
   connect(rotateRightAct, SIGNAL(triggered()), this, SLOT(rotateRight()));

   zoomInAct = new QAction(QIcon(":/new/prefix1/zoom-in.png"), tr("Zoom in"), this);
   zoomInAct->setStatusTip(tr("Zoom in graph"));
   connect(zoomInAct, SIGNAL(triggered()), this, SLOT(zoomIn()));

   zoomOutAct = new QAction(QIcon(":/new/prefix1/zoom-out.png"), tr("Zoom out"), this);
   zoomOutAct->setStatusTip(tr("Zoom out graph"));
   connect(zoomOutAct, SIGNAL(triggered()), this, SLOT(zoomOut()));

   zoomOrigAct = new QAction(QIcon(":/new/prefix1/zoom-original.png"), tr("Zoom original"), this);
   zoomOrigAct->setStatusTip(tr("Show graph using original size"));
   connect(zoomOrigAct, SIGNAL(triggered()), this, SLOT(zoomOriginal()));

   highlightNoneAct = new QAction(tr("None"), this);
   highlightNoneAct->setStatusTip(tr("Do not highlight sequence"));
   connect(highlightNoneAct, SIGNAL(triggered()), this, SLOT(highlightNone()));

   highlightFastaAAAct = new QAction(tr("FASTA protein"), this);
   highlightFastaAAAct->setStatusTip(tr("Highlight protein sequence in FASTA format"));
   connect(highlightFastaAAAct, SIGNAL(triggered()), this, SLOT(highlightFastaAA()));

   highlightFastaNucAct = new QAction(tr("FASTA nucleic acid"), this);
   highlightFastaNucAct->setStatusTip(tr("Highlight nucleic acid sequence in FASTA format"));
   connect(highlightFastaNucAct, SIGNAL(triggered()), this, SLOT(highlightFastaNuc()));

   highlightAlignmentAAAct = new QAction(tr("Alignment protein"), this);
   highlightAlignmentAAAct->setStatusTip(tr("Highlight protein sequence in alignment format"));
   connect(highlightAlignmentAAAct, SIGNAL(triggered()), this, SLOT(highlightAlignmentAA()));

   highlightAlignmentNucAct = new QAction(tr("Alignment nucleic acid"), this);
   highlightAlignmentNucAct->setStatusTip(tr("Highlight nucleic acid sequence in alignment format"));
   connect(highlightAlignmentNucAct, SIGNAL(triggered()), this, SLOT(highlightAlignmentNuc()));

   highlightEmblAct = new QAction(tr("EMBL"), this);
   highlightEmblAct->setStatusTip(tr("Highlight sequence in EMBL format"));
   connect(highlightEmblAct, SIGNAL(triggered()), this, SLOT(highlightEmbl()));

   highlightGbAAAct = new QAction(tr("Gb Protein"), this);
   highlightGbAAAct->setStatusTip(tr("Highlight protein sequence in GB format"));
   connect(highlightGbAAAct, SIGNAL(triggered()), this, SLOT(highlightGbAA()));

   highlightGbNucAct = new QAction(tr("Gb nuleic acid"), this);
   highlightGbNucAct->setStatusTip(tr("Highlight nucleic acid sequence in GB format"));
   connect(highlightGbNucAct, SIGNAL(triggered()), this, SLOT(highlightGbNuc()));

   helpAct = new QAction(QIcon(":/new/prefix1/help.png"), tr("EMBOSS Help"), this);
   helpAct->setShortcut(tr("Ctrl+?"));
   helpAct->setStatusTip(tr("Show help on EMBOSS program"));
   connect(helpAct, SIGNAL(triggered()), this, SLOT(embossHelp()));

   kembossWindow::createActions();
}

void kprogramWindow::createMenus()
{
   fileMenu = menuBar()->addMenu(tr("&File"));
   fileMenu->addAction(newAct);
   fileMenu->addAction(openAct);

   recentSettingsMenu = fileMenu->addMenu(QIcon(":/new/prefix1/open_recent.png"), tr("Recent Settings"));

   fileMenu->addMenu(recentSettingsMenu);
   fileMenu->addSeparator();
   fileMenu->addAction(saveAct);
   fileMenu->addAction(saveAsAct);
   fileMenu->addSeparator();
   fileMenu->addAction(propertiesAct);
   fileMenu->addAction(printAct);
   fileMenu->addSeparator();
   fileMenu->addAction(exitAct);

   editMenu = menuBar()->addMenu(tr("&Edit"));
   editMenu->addAction(cutAct);
   editMenu->addAction(copyAct);
   editMenu->addAction(pasteAct);
   editMenu->addSeparator();
   editMenu->addAction(undoAct);
   editMenu->addAction(redoAct);
   editMenu->addSeparator();

   graphMenu = editMenu->addMenu(tr("Graph"));
   graphMenu->addAction(rotateLeftAct);
   graphMenu->addAction(rotateRightAct);
   graphMenu->addSeparator();
   graphMenu->addAction(zoomOutAct);
   graphMenu->addAction(zoomOrigAct);
   graphMenu->addAction(zoomInAct);

   viewMenu = menuBar()->addMenu(tr("&View"));
  /* tabsMenu = viewMenu->addMenu(tr("Ta&bs"));
   QAction *tabAct;
   for (int i = 0; i < kpwMainTab->count(); ++i) {
      tabAct = new QAction(kpwMainTab->tabText(i),this);
      tabAct->setStatusTip(tr("Show ")+kpwMainTab->tabText(i)+tr(" tab"));
      tabAct->setData(i);
      tabAct->setCheckable(true);
      tabAct->setChecked(true);
      connect(tabAct, SIGNAL(triggered()), this, SLOT(viewTab()));
      tabsMenu->addAction(tabAct);
   }*/
   syntaxHighlightingMenu = viewMenu->addMenu(tr("S&yntax highlighting"));
   syntaxHighlightingMenu->addAction(highlightNoneAct);
   syntaxHighlightingMenu->addAction(highlightFastaAAAct);
   syntaxHighlightingMenu->addAction(highlightFastaNucAct);
   syntaxHighlightingMenu->addAction(highlightAlignmentAAAct);
   syntaxHighlightingMenu->addAction(highlightAlignmentNucAct);
   syntaxHighlightingMenu->addAction(highlightEmblAct);
   syntaxHighlightingMenu->addAction(highlightGbAAAct);
   syntaxHighlightingMenu->addAction(highlightGbNucAct);

   toolbarsMenu = viewMenu->addMenu(tr("&Toolbars"));
   toolbarsMenu->addAction(viewFileToolbarAct);
   toolbarsMenu->addAction(viewEditToolbarAct);
   toolbarsMenu->addAction(viewViewToolbarAct);
   toolbarsMenu->addAction(viewGraphToolbarAct);
   toolbarsMenu->addAction(viewRunToolbarAct);
   toolbarsMenu->addAction(viewHelpToolbarAct);

   runMenu = menuBar()->addMenu(tr("&Run"));
   runMenu->addAction(addParameterAct);
   runMenu->addAction(copyCommandlineAct);
   runMenu->addAction(runAct);

   menuBar()->addSeparator();  //This has no effect for most widget styles (e.g., Windows and Mac OS X styles), but for Motif-based styles this makes sure that Help is pushed to the right side of the menu bar.

   helpMenu = menuBar()->addMenu(tr("&Help"));
   helpMenu->addAction(helpAct);
   helpMenu->addAction(reportBugAct);
   helpMenu->addAction(aboutAct);
}

void kprogramWindow::createToolbars()
{
   fileToolBar = addToolBar(tr("File"));
   fileToolBar->addAction(exitAct);
   fileToolBar->addAction(newAct);
   fileToolBar->addAction(openAct);
   fileToolBar->addAction(saveAct);
   fileToolBar->addAction(saveAsAct);
   fileToolBar->addSeparator();
   fileToolBar->addAction(printAct);

   editToolBar = addToolBar(tr("Edit"));
   editToolBar->addAction(cutAct);
   editToolBar->addAction(copyAct);
   editToolBar->addAction(pasteAct);
   editToolBar->addSeparator();
   editToolBar->addAction(undoAct);
   editToolBar->addAction(redoAct);

   viewToolBar = addToolBar(tr("View"));
   selectHighlighting = new QComboBox();
   QStringList highlightingPatterns;
   highlightingPatterns << "none" << "FASTA Protein" << "FASTA nucleic acid" << "Alignment protein" << "Alignment nucleic acid" << "EMBL" << "Genebank protein" << "Genebank (full)";
   selectHighlighting->addItems(highlightingPatterns);
   connect(selectHighlighting, SIGNAL(currentIndexChanged(int)), this, SLOT(highlightingChanges(int)));
   viewToolBar->addWidget(selectHighlighting);

   graphToolBar = addToolBar(tr("Graph"));
   graphToolBar->addAction(rotateLeftAct);
   graphToolBar->addAction(rotateRightAct);
   graphToolBar->addSeparator();
   graphToolBar->addAction(zoomOutAct);
   graphToolBar->addAction(zoomOrigAct);
   graphToolBar->addAction(zoomInAct);

   runToolBar = addToolBar(tr("Run"));
   runToolBar->addAction(addParameterAct);
   runToolBar->addAction(copyCommandlineAct);
   runToolBar->addAction(runAct);

   helpToolBar = addToolBar(tr("Help"));
   helpToolBar->addAction(helpAct);
}

/*void kprogramWindow::viewTab()
{
   QAction *action = qobject_cast<QAction *>(sender());
   if (action) {
      kpwMainTab->setTabEnabled(action->data().toInt(), !action->isChecked());
      action->setChecked(!action->isChecked());
   }
}*/

void kprogramWindow::findControls()
{
   QList<QObject *> allObjects = findChildren<QObject *>();
   foreach (QObject *object, allObjects) {
    if (object->objectName() != "" && object->objectName() != "qt_spinbox_lineedit" && object->objectName() != "qt_menubar_ext_button" && object->objectName() != "qt_toolbar_ext_button") {
      QLineEdit *lineEdit = qobject_cast<QLineEdit *>(object);
      if (lineEdit) {
         connect(lineEdit, SIGNAL(textEdited(QString)), this, SLOT(wasModified()));
         if (lineEdit->objectName() == "InputFileLineEdit") kpwInputFileLineEdit = lineEdit;
         if (lineEdit->objectName() == "OutputFileLineEdit") kpwOutputFileLineEdit = lineEdit;
      }
      QComboBox *comboBox = qobject_cast<QComboBox *>(object);
      if (comboBox) {
         connect(comboBox, SIGNAL(activated(int)), this, SLOT(wasModified()));
         if (comboBox->objectName() == "ProgramComboBox") kpwProgramComboBox = comboBox;
         if (comboBox->objectName() == "GeneticCodeComboBox") loadFormats("genetic_codes", comboBox);
         if (comboBox->objectName() == "FormatComboBox") loadFormats("graph_types", comboBox);
         if (comboBox->objectName() == "SubstitutionMatrixComboBox") loadFormats("substitution_matrices", comboBox);
         if (comboBox->objectName() == "CodonUsageFileComboBox") loadFormats("codon_usage_files", comboBox);
      }
      QSpinBox *spinBox = qobject_cast<QSpinBox *>(object);
      if (spinBox) connect(spinBox, SIGNAL(valueChanged(int)), this, SLOT(wasModified()));
      QDoubleSpinBox *doubleSpinBox = qobject_cast<QDoubleSpinBox *>(object);
      if (doubleSpinBox) connect(doubleSpinBox, SIGNAL(valueChanged(double)), this, SLOT(wasModified()));
      QAbstractButton *abstractButton = qobject_cast<QAbstractButton *>(object);
      if (abstractButton && !abstractButton->objectName().isEmpty() && !abstractButton->text().isEmpty()) {
         connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(wasModified()));
         if (abstractButton->objectName() =="SequenceAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openSAQForm()));
         if (abstractButton->objectName() =="OutseqAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openOsAQForm()));
         if (abstractButton->objectName() =="OutfeatAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openOfAQForm()));
         if (abstractButton->objectName() =="GraphAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openGAQForm()));
         if (abstractButton->objectName() =="AlignmentAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openAAQForm()));
         if (abstractButton->objectName() =="OutfileAssociatedQualifiersPushButton") connect(abstractButton, SIGNAL(clicked(bool)), this, SLOT(openOAQForm()));
      }
      QGroupBox *groupBox = qobject_cast<QGroupBox *>(object);
      if (groupBox && groupBox->isCheckable()) connect(groupBox, SIGNAL(clicked(bool)), this, SLOT(wasModified()));

      QTextEdit *textEdit = qobject_cast<QTextEdit *>(object);
      if (textEdit) {
         if (textEdit->objectName() == "InputTextEdit") {
            kpwInputTextEdit = textEdit;
            kpwInputTextEdit->setAcceptRichText(false);
         }
         if (textEdit->objectName() == "OutputTextEdit") {
            kpwOutputTextEdit = textEdit;
            kpwOutputTextEdit->setAcceptRichText(false);
         }
      }
      QGraphicsView *graphicsView = qobject_cast<QGraphicsView *>(object);
      if (graphicsView && graphicsView->objectName() == "OutputGraphicsView") {
         kpwOutputGraphicsView = graphicsView;
      }
      QTabWidget *tabWidget = qobject_cast<QTabWidget *>(object);
      if (tabWidget && tabWidget->objectName() == "MainTab") {
         kpwMainTab = tabWidget;
         connect(kpwMainTab, SIGNAL(currentChanged(int)), this, SLOT(tabChanged(int)));
      }
    }
   }
}

void kprogramWindow::tabChanged(const int &index)
{
   QTextEdit *currentTextEdit = NULL;
   if (kpwInputTextEdit != NULL) {
      disconnect(kpwInputTextEdit, SIGNAL(copyAvailable(bool)), 0, 0);
      disconnect(kpwInputTextEdit, SIGNAL(undoAvailable(bool)), 0, 0);
      disconnect(kpwInputTextEdit, SIGNAL(redoAvailable(bool)), 0, 0);
      if (kpwMainTab->tabText(index) == "Input") currentTextEdit = kpwInputTextEdit;
   }
   if (kpwOutputTextEdit != NULL) {
      disconnect(kpwOutputTextEdit, SIGNAL(copyAvailable(bool)), 0, 0);
      disconnect(kpwOutputTextEdit, SIGNAL(undoAvailable(bool)), 0, 0);
      disconnect(kpwOutputTextEdit, SIGNAL(redoAvailable(bool)), 0, 0);
      if (kpwMainTab->tabText(index) == "Output") currentTextEdit = kpwOutputTextEdit;
   }
   disconnect(cutAct, SIGNAL(triggered()), 0, 0);
   disconnect(copyAct, SIGNAL(triggered()), 0, 0);
   disconnect(pasteAct, SIGNAL(triggered()), 0, 0);
   disconnect(undoAct, SIGNAL(triggered()), 0, 0);
   disconnect(redoAct, SIGNAL(triggered()), 0, 0);

   if (currentTextEdit != NULL) {
      cutAct->setEnabled(!currentTextEdit->textCursor().selectedText().isEmpty());
      connect(currentTextEdit, SIGNAL(copyAvailable(bool)), cutAct, SLOT(setEnabled(bool)));
      copyAct->setEnabled(!currentTextEdit->textCursor().selectedText().isEmpty());
      connect(currentTextEdit, SIGNAL(copyAvailable(bool)), copyAct, SLOT(setEnabled(bool)));
      connect(cutAct, SIGNAL(triggered()), currentTextEdit, SLOT(cut()));
      connect(copyAct, SIGNAL(triggered()), currentTextEdit, SLOT(copy()));
      connect(pasteAct, SIGNAL(triggered()), currentTextEdit, SLOT(paste()));
      undoAct->setEnabled(currentTextEdit->document()->isUndoAvailable());
      connect(currentTextEdit, SIGNAL(undoAvailable(bool)), undoAct, SLOT(setEnabled(bool)));
      redoAct->setEnabled(currentTextEdit->document()->isRedoAvailable());
      connect(currentTextEdit, SIGNAL(redoAvailable(bool)), redoAct, SLOT(setEnabled(bool)));
      connect(undoAct, SIGNAL(triggered()), currentTextEdit, SLOT(undo()));
      connect(redoAct, SIGNAL(triggered()), currentTextEdit, SLOT(redo()));
      printAct->setEnabled(true);
      highlighter->setDocument(currentTextEdit->document());
      selectHighlighting->setEnabled(true);
      if (currentTextEdit == kpwInputTextEdit) {
         selectHighlighting->setCurrentIndex(highlightInput);
         highlighter->setSyntax(highlightInput);
      } else {
         selectHighlighting->setCurrentIndex(highlightOutput);
         highlighter->setSyntax(highlightOutput);
      }
   } else {
      cutAct->setEnabled(false);
      copyAct->setEnabled(false);
      undoAct->setEnabled(false);
      redoAct->setEnabled(false);
      printAct->setEnabled(false);
      selectHighlighting->setEnabled(false);
   }
   if (kpwOutputGraphicsView != NULL && kpwMainTab->tabText(index) == "Output") {
      graphMenu->setEnabled(true);
      graphToolBar->setEnabled(true);
      printAct->setEnabled(true);
   } else {
      graphMenu->setEnabled(false);
      graphToolBar->setEnabled(false);
   }
}

void kprogramWindow::setTextEditFont()
{
   QTextCursor currentCursor;
   if (kpwInputTextEdit != NULL) {
      currentCursor = kpwInputTextEdit->textCursor();
      kpwInputTextEdit->selectAll();
      kpwInputTextEdit->setCurrentFont(qobject_cast<kemboss *>(kembossParent)->ksettings.textEditFont());
      kpwInputTextEdit->document()->setDefaultFont(qobject_cast<kemboss *>(kembossParent)->ksettings.textEditFont());
      if (qobject_cast<kemboss *>(kembossParent)->ksettings.wrapLines()) kpwInputTextEdit->setLineWrapMode(QTextEdit::WidgetWidth);
      else kpwInputTextEdit->setLineWrapMode(QTextEdit::NoWrap);
      kpwInputTextEdit->setTextCursor(currentCursor);
   }
   if (kpwOutputTextEdit != NULL) {
      currentCursor = kpwOutputTextEdit->textCursor();
      kpwOutputTextEdit->selectAll();
      kpwOutputTextEdit->selectAll();
      kpwOutputTextEdit->setCurrentFont(qobject_cast<kemboss *>(kembossParent)->ksettings.textEditFont());
      kpwOutputTextEdit->document()->setDefaultFont(qobject_cast<kemboss *>(kembossParent)->ksettings.textEditFont());
      if (qobject_cast<kemboss *>(kembossParent)->ksettings.wrapLines()) kpwOutputTextEdit->setLineWrapMode(QTextEdit::WidgetWidth);
      else kpwOutputTextEdit->setLineWrapMode(QTextEdit::NoWrap);
      kpwOutputTextEdit->setTextCursor(currentCursor);
   }
}

void kprogramWindow::loadFormats(const QString &formatFile, QWidget *widget)
{
   QFile file(qobject_cast<kemboss *>(kembossParent)->ksettings.kembossData() + formatFile);
   if (file.open(QIODevice::ReadOnly | QIODevice::Text)) {
      QTextStream in(&file);
      while (!in.atEnd()) {
         QString line = in.readLine();
         qobject_cast<QComboBox *>(widget)->addItem(line);
      }
   }
}

void kprogramWindow::newSettings()
{
   if (maybeSave()) {
      deleteTemporaryFiles();
      if (kpwInputTextEdit != NULL) {
         kpwInputTextEdit->clear();
         kpwInputTextEdit->document()->setModified(false);
      }
      if (kpwOutputTextEdit != NULL) {
         kpwOutputTextEdit->clear();
      }
      // Set DefaultValues
      if (QFile::exists(defaultSettingsFile()))
         loadFile(defaultSettingsFile());
      setCurrentFile("");
   }
}

bool kprogramWindow::loadFile(const QString &fileName)
{
   statusBar()->showMessage(tr("Opening settings"));
   QFile file(fileName);
   if (!file.open(QFile::ReadOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                    tr("Cannot read file %1:\n%2.")
                    .arg(fileName)
                    .arg(file.errorString()));
      return false;
   }

   // Load Settings
   QSettings settings(fileName, QSettings::IniFormat);

   QList<QWidget *> allWidgets = findChildren<QWidget *>();
   foreach (QWidget *widget, allWidgets) {
    if (widget->objectName() != "" && widget->objectName() != "qt_spinbox_lineedit" && widget->objectName() != "qt_menubar_ext_button" && widget->objectName() != "qt_toolbar_ext_button" && widget->objectName() !="ProgramComboBox") {
      QLineEdit *lineEdit = qobject_cast<QLineEdit *>(widget);
      if (lineEdit)
         lineEdit->setText(settings.value(lineEdit->objectName()).toString());
      QComboBox *comboBox = qobject_cast<QComboBox *>(widget);
      if (comboBox)
         comboBox->setCurrentIndex(comboBox->findText(settings.value(comboBox->objectName()).toString()));
      QSpinBox *spinBox = qobject_cast<QSpinBox *>(widget);
      if (spinBox)
         spinBox->setValue(settings.value(spinBox->objectName()).toInt());
      QDoubleSpinBox *doubleSpinBox = qobject_cast<QDoubleSpinBox *>(widget);
      if (doubleSpinBox)
         doubleSpinBox->setValue(settings.value(doubleSpinBox->objectName()).toDouble());
      QAbstractButton *abstractButton = qobject_cast<QAbstractButton *>(widget);
      if (abstractButton && abstractButton->isCheckable())
         abstractButton->setChecked(settings.value(abstractButton->objectName()).toBool());
      QGroupBox *groupBox = qobject_cast<QGroupBox *>(widget);
      if (groupBox && groupBox->isCheckable())
         groupBox->setChecked(settings.value(groupBox->objectName()).toBool());
    }
   }
   associatedQualifiers.clear();
   associatedQualifiers+=settings.value("associated_qualifiers").toStringList();
   additionalConfigurationParameters.clear();
   additionalConfigurationParameters+=settings.value("additional_configuration_parameters").toStringList();
   statusBar()->clearMessage();

   // Open input file if specified
   if (kpwInputFileLineEdit != NULL && kpwInputFileLineEdit->text() != "")
      loadInputFile(kpwInputFileLineEdit->text());

   return true;
}

bool kprogramWindow::saveFile(const QString &fileName)
{
   statusBar()->showMessage(tr("Saving settings"));
   QFile file(fileName);
   if (!file.open(QFile::WriteOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                   tr("Cannot write file %1:\n%2.")
                   .arg(fileName)
                   .arg(file.errorString()));
      return false;
   }

   // Save Settings
   QSettings settings(fileName, QSettings::IniFormat);
   settings.setValue("kemboss_version",KEMBOSS_VERSION);

   QList<QWidget *> allWidgets = findChildren<QWidget *>();
   foreach (QWidget *widget, allWidgets) {
    if (widget->objectName() != "" && widget->objectName() != "qt_spinbox_lineedit" && widget->objectName() != "qt_menubar_ext_button" && widget->objectName() != "qt_toolbar_ext_button") {
      QLineEdit *lineEdit = qobject_cast<QLineEdit *>(widget);
      if (lineEdit) settings.setValue(lineEdit->objectName(),lineEdit->text());
      QComboBox *comboBox = qobject_cast<QComboBox *>(widget);
      if (comboBox) settings.setValue(comboBox->objectName(),comboBox->currentText());
      QSpinBox *spinBox = qobject_cast<QSpinBox *>(widget);
      if (spinBox) settings.setValue(spinBox->objectName(),spinBox->value());
      QDoubleSpinBox *doubleSpinBox = qobject_cast<QDoubleSpinBox *>(widget);
      if (doubleSpinBox) settings.setValue(doubleSpinBox->objectName(),doubleSpinBox->value());
      QAbstractButton *abstractButton = qobject_cast<QAbstractButton *>(widget);
      if (abstractButton && abstractButton->isCheckable()) settings.setValue(abstractButton->objectName(),abstractButton->isChecked());
      QGroupBox *groupBox = qobject_cast<QGroupBox *>(widget);
      if (groupBox && groupBox->isCheckable()) settings.setValue(groupBox->objectName(),groupBox->isChecked());
    }
   }
   if (!associatedQualifiers.isEmpty()) settings.setValue("associated_qualifiers",associatedQualifiers);
   if (!additionalConfigurationParameters.isEmpty()) settings.setValue("additional_configuration_parameters",additionalConfigurationParameters);
   setCurrentFile(fileName);
   statusBar()->clearMessage();
   return true;
}

void kprogramWindow::setCurrentFile(const QString &fileName)
{
   if (fileName == defaultSettingsFile()) curFile.clear();
   else curFile = fileName;
   setWindowModified(false); saveAct->setEnabled(false);
   QString shownName;
   if (curFile.isEmpty())
      shownName = "untitled."+objectName();
   else {
      shownName = strippedName(curFile);
      QSettings settings("kemboss.org", objectName());
      QStringList files = settings.value("recentFileList").toStringList();
      files.removeAll(fileName);
      files.prepend(fileName);
      while (files.size() > MaxRecentFiles)
         files.removeLast();
      settings.setValue("recentFileList", files);

      foreach (kembossWindow *window, qobject_cast<kemboss *>(kembossParent)->programWindows) {
         if (window->objectName() == objectName())
            window->updateRecentFileActions();

      qobject_cast<kemboss *>(kembossParent)->ksettings.setcurrentDir(QFileInfo(fileName).absolutePath());
      }
   }
   setWindowTitle(tr("%1[*] - %2").arg(shownName).arg(objectName()));
}

bool kprogramWindow::on_OpenInputFilePushButton_clicked()
{
   if (!maybeSaveInputFile())
      return false;
   QString fileName = QFileDialog::getOpenFileName(this, tr("Open input file"),qobject_cast<kemboss *>(kembossParent)->ksettings.currentDir(),"All files (*)");
   if (!fileName.isEmpty()) {
      if (kpwInputFileLineEdit != NULL)
         kpwInputFileLineEdit->setText(fileName);
      qobject_cast<kemboss *>(kembossParent)->ksettings.setcurrentDir(QFileInfo(fileName).absolutePath());
      loadInputFile(fileName);
   } else return false;
   return true;
}

bool kprogramWindow::loadInputFile(const QString &fileName)
{
   if (kpwInputTextEdit == NULL) return false;
   statusBar()->showMessage(tr("Opening input file"));
   QFile file(fileName);
   if (!file.open(QFile::ReadOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                           tr("Cannot open input file %1:\n%2.")
                           .arg(fileName)
                           .arg(file.errorString()));
      statusBar()->clearMessage();
      return false;
   }
   QString suffix = QFileInfo(fileName).suffix();
   if (suffix == "fasta")
      selectHighlighting->setCurrentIndex(1);
   else if ((suffix == "alignment") || (suffix == "align") || (suffix == "aln"))
      selectHighlighting->setCurrentIndex(3);
   else if ((suffix == "embl") || (suffix == "sp") || (suffix == "swissprot") || (suffix == "swiss"))
      selectHighlighting->setCurrentIndex(5);
   else if ((suffix == "gb") || (suffix == "genebank"))
      selectHighlighting->setCurrentIndex(6);
   else
      selectHighlighting->setCurrentIndex(0);
   QTextStream in(&file);
   kpwInputTextEdit->setPlainText(in.readAll());
   kpwInputTextEdit->document()->setModified(false);
   statusBar()->clearMessage();
   return true;
}

//void kprogramWindow::appendText(QString text, bool finished)
//{
//   activeTextEdit->append(text);
//  if (finished)
//      activeTextEdit->document()->setModified(false);
//}

bool kprogramWindow::on_SaveInputFileAsPushButton_clicked()
{
   if (kpwInputTextEdit != NULL) {
      QString fileName = QFileDialog::getSaveFileName(this, tr("Save input file"),qobject_cast<kemboss *>(kembossParent)->ksettings.currentDir(),"All files (*)");
      if (!fileName.isEmpty() && saveInputFile(fileName)) {
         if (kpwInputFileLineEdit != NULL) kpwInputFileLineEdit->setText(fileName);
         qobject_cast<kemboss *>(kembossParent)->ksettings.setcurrentDir(QFileInfo(fileName).absolutePath());
      } else return false;
   }
   return true;
}

bool kprogramWindow::saveInputFile(const QString &fileName)
{
   statusBar()->showMessage(tr("Saving input file"));
   QFile file(fileName);
   if (!file.open(QFile::WriteOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                           tr("Cannot save input file %1:\n%2.")
                           .arg(fileName)
                           .arg(file.errorString()));
      statusBar()->clearMessage();
      return false;
   }
   QTextStream out(&file);
   out << kpwInputTextEdit->toPlainText();
   kpwInputTextEdit->document()->setModified(false);
   statusBar()->clearMessage();
   return true;
}

bool kprogramWindow::on_SaveOutputFileAsPushButton_clicked()
{
   QString fileName = QFileDialog::getSaveFileName(this, tr("Save output"),qobject_cast<kemboss *>(kembossParent)->ksettings.currentDir(),"All files (*)");
   if (fileName.isEmpty()) return false;

   if (kpwOutputTextEdit != NULL && kpwOutputTextEdit->document()->isModified()) {
   // The contents of the output text edit as been modified by the user
      saveOutputFile(fileName);
   } else if (!tempOutputFile.isEmpty() && QFile::exists(tempOutputFile)) {
   // The output of the EMBOSS program has been saved to a temporary file but has not been changed by the user
      if (QFile::exists(fileName))
         QFile::remove(fileName);
      QFile::rename(tempOutputFile,fileName);
      tempOutputFile.clear();
   }
   if (kpwOutputFileLineEdit != NULL)
      kpwOutputFileLineEdit->setText(fileName);
   qobject_cast<kemboss *>(kembossParent)->ksettings.setcurrentDir(QFileInfo(fileName).absolutePath());
   return true;
}

bool kprogramWindow::saveOutputFile(const QString &fileName)
{
   statusBar()->showMessage(tr("Saving output file"));
   QFile file(fileName);
   if (!file.open(QFile::WriteOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                           tr("Cannot save output file %1:\n%2.")
                           .arg(fileName)
                           .arg(file.errorString()));
      statusBar()->clearMessage();
      return false;
   }
   QTextStream out(&file);
   out << kpwOutputTextEdit->toPlainText();
   kpwOutputTextEdit->document()->setModified(false);
   if (QFile::exists(tempOutputFile)) {
   // If a temporary EMBOSS output file exists, it has to be deleted
      QFile::remove(tempOutputFile);
      tempOutputFile.clear();
   }
   statusBar()->clearMessage();
   return true;
}

void kprogramWindow::loadPrinterSettings()
{
   // Load printer settings
   QSettings settings("kemboss.org", "Configuration");
   printer.setCollateCopies(settings.value("print_collate_copies").toBool());
   printer.setColorMode(QPrinter::ColorMode(settings.value("print_color_mode").toInt()));
   printer.setDuplex(QPrinter::DuplexMode(settings.value("print_duplex_mode").toInt()));
   printer.setCreator(objectName());
   printer.setFontEmbeddingEnabled(settings.value("print_font_embedding_enabled").toBool());
   printer.setOrientation(QPrinter::Orientation(settings.value("print_orientation").toInt()));
   printer.setOutputFormat(QPrinter::OutputFormat(settings.value("print_output_format").toInt()));
   printer.setPageMargins(settings.value("print_margin_left").toDouble(),
                          settings.value("print_margin_top").toDouble(),
                          settings.value("print_margin_right").toDouble(),
                          settings.value("print_margin_bottom").toDouble(),
                          QPrinter::Unit(settings.value("print_unit").toInt()));
   printer.setPageOrder(QPrinter::PageOrder(settings.value("print_page_order").toInt()));
   printer.setPaperSize(QPrinter::PaperSize(settings.value("print_paper_size").toInt()));
   printer.setPaperSource(QPrinter::PaperSource(settings.value("print_paper_source").toInt()));
   printer.setPrinterName(settings.value("print_printer_name").toString());
   printer.setResolution(settings.value("print_resolution").toInt());
}

void kprogramWindow::savePrinterSettings()
{
   // Save the current settings of the printer object
   QSettings settings("kemboss.org", "Configuration");
   settings.setValue("print_collate_copies", printer.collateCopies());
   settings.setValue("print_color_mode", printer.colorMode());
   settings.setValue("print_duplex_mode", printer.duplex());
   settings.setValue("print_font_embedding_enabled", printer.fontEmbeddingEnabled());
   qreal pleft, ptop, pright, pbottom;
   QPrinter::Unit unit = QPrinter::Millimeter;
   printer.getPageMargins(&pleft, &ptop, &pright, &pbottom, unit);
   settings.setValue("print_margin_left", pleft);
   settings.setValue("print_margin_top", ptop);
   settings.setValue("print_margin_right", pright);
   settings.setValue("print_margin_bottom", pbottom);
   settings.setValue("print_unit", unit);
   settings.setValue("print_orientation", printer.orientation());
   settings.setValue("print_output_format", printer.outputFormat());
   settings.setValue("print_page_order", printer.pageOrder());
   settings.setValue("print_paper_size", printer.paperSize());
   settings.setValue("print_paper_source", printer.paperSource());
   settings.setValue("print_printer_name", printer.printerName());
   settings.setValue("print_resolution", printer.resolution());
}

void kprogramWindow::printDocument()
{
#ifndef QT_NO_PRINTER
   QTextEdit *currentTextEdit = NULL;
   if (kpwMainTab->tabText(kpwMainTab->currentIndex()) == "Input" && kpwInputTextEdit != NULL) currentTextEdit = kpwInputTextEdit;
   else if (kpwMainTab->tabText(kpwMainTab->currentIndex()) == "Output" && kpwOutputTextEdit != NULL) currentTextEdit = kpwOutputTextEdit;

   QPrintDialog *dialog = new QPrintDialog(&printer, this);
   dialog->setWindowTitle(tr("Print Document"));
   if (currentTextEdit != NULL && currentTextEdit->textCursor().hasSelection())
      dialog->addEnabledOption(QAbstractPrintDialog::PrintSelection);
   if (dialog->exec() != QDialog::Accepted)
      return;
   savePrinterSettings();

   if (currentTextEdit != NULL)
      currentTextEdit->print(&printer);
   else if (kpwMainTab->tabText(kpwMainTab->currentIndex()) == "Output" && kpwOutputGraphicsView != NULL) {
      QPainter painter(&printer);
      kpwOutputGraphicsView->render(&painter);
   }
#else
   QMessageBox::information(this, tr("Printing canceled"), tr("Printing is not supported on this Qt build"), QMessageBox::Cancel);
#endif
}

void kprogramWindow::documentProperties()
{
   QPageSetupDialog *dialog = new QPageSetupDialog(&printer, this);
   if (dialog->exec() == QDialog::Accepted)
      savePrinterSettings();
}

void kprogramWindow::highlightNone()
{
   selectHighlighting->setCurrentIndex(0);
}

void kprogramWindow::highlightFastaAA()
{
   selectHighlighting->setCurrentIndex(1);
}

void kprogramWindow::highlightFastaNuc()
{
   selectHighlighting->setCurrentIndex(2);
}

void kprogramWindow::highlightAlignmentAA()
{
   selectHighlighting->setCurrentIndex(3);
}

void kprogramWindow::highlightAlignmentNuc()
{
   selectHighlighting->setCurrentIndex(4);
}

void kprogramWindow::highlightEmbl()
{
   selectHighlighting->setCurrentIndex(5);
}

void kprogramWindow::highlightGbAA()
{
   selectHighlighting->setCurrentIndex(6);
}

void kprogramWindow::highlightGbNuc()
{
   selectHighlighting->setCurrentIndex(7);
}

void kprogramWindow::highlightingChanges(int index)
{
   if (kpwMainTab->tabText(kpwMainTab->currentIndex()) == "Input") {
      highlightInput = index;
   } else if (kpwMainTab->tabText(kpwMainTab->currentIndex()) == "Output") {
      highlightOutput = index;
   }
   highlighter->setSyntax(index);
   highlighter->rehighlight();
}

void kprogramWindow::zoomIn()
{
   kpwOutputGraphicsView->scale(1.2, 1.2);
   graphScale *= 1.2;
}

void kprogramWindow::zoomOut()
{
   kpwOutputGraphicsView->scale(0.8, 0.8);
   graphScale *= 0.8;
}

void kprogramWindow::zoomOriginal()
{
   kpwOutputGraphicsView->scale(1/graphScale, 1/graphScale);
   graphScale = 1;
}

void kprogramWindow::rotateLeft()
{
   kpwOutputGraphicsView->rotate(-90);
}

void kprogramWindow::rotateRight()
{
   kpwOutputGraphicsView->rotate(90);
}


bool kprogramWindow::runEmbossApp() {
   deleteTemporaryFiles();
   if (!checkInput()) return false;
   if (!checkOutput()) return false;
   if (!createProgramConfiguration()) return false;
   beforeRunningProgram();
   if (!checkEmbossApp()) return false;

   // Create program progress window
   if (programProgress == NULL) {
      programProgress = new programProgressForm(&qobject_cast<kemboss *>(kembossParent)->ksettings);
      connect(programProgress->CancelButton, SIGNAL(clicked()), this, SLOT(killProcess()));
   }
   programProgress->ProgramOutputTextEdit->document()->clear();
   programProgress->move(x()+100,y()+150);
   programProgress->show();

   // Display command line options for EMBOSS application in program progress window
   programProgress->ProgramOutputTextEdit->append("Running "+embossApp());
   programProgress->ProgramOutputTextEdit->append(programConfiguration.join("\n"));

   runTime.start();
   programProcess.start(qobject_cast<kemboss *>(kembossParent)->ksettings.embossBin()+embossApp()
   #ifdef Q_WS_WIN
   +".exe"
   #endif
   , programConfiguration);

   return true;
}

bool kprogramWindow::checkInput()
{
   if (kpwInputFileLineEdit != NULL) {
      // If no input file is specified but text has been entered in the InputFileTextEdit, create a temporary input file.
      if (kpwInputFileLineEdit->text().isEmpty()) {
         if (kpwInputTextEdit == NULL || kpwInputTextEdit->document()->isEmpty()) {
            QMessageBox::warning(this, objectName(),
                   tr("No input available!\n"
                   "Please, either select an appropriate input file or enter Your input data into the corresponding text box."));
            return false;
         }
         #ifdef Q_WS_WIN
         tempInputFile = QDir::tempPath();
         #else
         tempInputFile.clear();
         #endif
         tempInputFile += tmpnam(NULL);
         kpwInputFileLineEdit->setText(tempInputFile);
         saveInputFile(tempInputFile);
         kpwInputTextEdit->document()->setModified();
      // If the text in the InputFileTextEdit has changed since the input file has been opened, ask what to do
      } else if (kpwInputTextEdit != NULL && kpwInputTextEdit->document()->isModified()) {
         QMessageBox::StandardButton ret;
         ret = QMessageBox::warning(this, objectName(),
              tr("The input file has been modified.\n"
              "Do you want to save your changes before running the emboss application?"),
              QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
         if (ret == QMessageBox::Save) {
            if (!saveInputFile(kpwInputFileLineEdit->text()))
               return false;
         } else if (ret == QMessageBox::Cancel) {
            return false;
         }
      }
   }
   return true;
}

bool kprogramWindow::checkOutput()
{
   if (kpwOutputFileLineEdit != NULL) {

      // If the output file is not specified create a temporary file
      if (kpwOutputFileLineEdit->text().isEmpty()) {
         #ifdef Q_WS_WIN
         tempOutputFile = QDir::tempPath();
         #else
         tempOutputFile.clear();
         #endif
         tempOutputFile += tmpnam(NULL);
         kpwOutputFileLineEdit->setText(tempOutputFile);

      // Delete file if output file already exists and auto-delete option is enabled. Otherwise ask what to do
      } else {
         if (QFile::exists(kpwOutputFileLineEdit->text())) {
            if (qobject_cast<kemboss *>(kembossParent)->ksettings.autoDeleteExistingOutputFile())
               QFile::remove(kpwOutputFileLineEdit->text());
            else {
               QMessageBox::StandardButton ret;
               ret = QMessageBox::warning(this, objectName(),
                  tr("The output file already exists.\n"
                  "Delete current output file?"),
                  QMessageBox::Yes | QMessageBox::Ignore | QMessageBox::Cancel);
               if (ret == QMessageBox::Yes)
                  QFile::remove(kpwOutputFileLineEdit->text());
               else if (ret == QMessageBox::Cancel)
                  return false;
            }
         }
      }
   }
   if (kpwOutputTextEdit != NULL) kpwOutputTextEdit->clear();
   return true;
}

bool kprogramWindow::checkEmbossApp()
{
   // Check wether EMBOSS application exists
   if (!QFile::exists(qobject_cast<kemboss *>(kembossParent)->ksettings.embossBin()+embossApp()
   #ifdef Q_WS_WIN
   +".exe"
   #endif
   )) {
      QMessageBox::warning(this, objectName(),
                   tr("EMBOSS application ")+embossApp()+tr(" couldn't be found!\n"
                   "Please check Your kemboss settings."));
      return false;
   }
   return true;
}

bool kprogramWindow::createProgramConfiguration()
{
   programConfiguration.clear();
   programConfiguration.append("-auto");

   QList<QWidget *> allWidgets = findChildren<QWidget *>();
   foreach (QWidget *widget, allWidgets) {
    if (!widget->toolTip().isEmpty() && widget->isEnabled()) {
      QLineEdit *lineEdit = qobject_cast<QLineEdit *>(widget);
      if (lineEdit && !lineEdit->text().isEmpty())
         programConfiguration.append("-"+lineEdit->toolTip()+"="+lineEdit->text());
      QComboBox *comboBox = qobject_cast<QComboBox *>(widget);
      if (comboBox) {
         if (!(comboBox->currentText()=="default"))
            programConfiguration.append("-"+comboBox->toolTip()+"="+comboBox->currentText());
      }
      QSpinBox *spinBox = qobject_cast<QSpinBox *>(widget);
      if (spinBox) programConfiguration.append("-"+spinBox->toolTip()+"="+spinBox->text());
      QDoubleSpinBox *doubleSpinBox = qobject_cast<QDoubleSpinBox *>(widget);
      if (doubleSpinBox) {
         QString num;
         num.setNum(doubleSpinBox->value());
         programConfiguration.append("-"+doubleSpinBox->toolTip()+"="+num);
      }
      QAbstractButton *abstractButton = qobject_cast<QAbstractButton *>(widget);
      if (abstractButton && abstractButton->isCheckable()) {
         if (abstractButton->toolTip().startsWith("[no]")) {
            if (abstractButton->isChecked())
               programConfiguration.append("-"+abstractButton->toolTip().remove(0,4));
            else programConfiguration.append("-"+abstractButton->toolTip().replace(0,4,"no"));
         } else {
            if (abstractButton->isChecked())
               programConfiguration.append("-"+abstractButton->toolTip());
         }
      }
      QGroupBox *groupBox = qobject_cast<QGroupBox *>(widget);
      if (groupBox && groupBox->isCheckable()) {
         if (groupBox->toolTip().startsWith("[no]")) {
            if (!groupBox->isChecked())
               programConfiguration.append("-"+groupBox->toolTip().replace(0,4,"no"));
         } else {
            if (groupBox->isChecked())
               programConfiguration.append("-"+groupBox->toolTip());
         }
      }
    }
   }
   programConfiguration += associatedQualifiers;
   programConfiguration += additionalConfigurationParameters;
   return true;
}

void kprogramWindow::beforeRunningProgram()
{

}

void kprogramWindow::completeOutputFiles(QLineEdit *lineEdit, const QString &extension)
{
   if (lineEdit->text().isEmpty()) {
      lineEdit->setText(kpwOutputFileLineEdit->text()+"."+extension);
      programConfiguration.append("-"+lineEdit->toolTip()+"="+lineEdit->text());
      // Temporary output files have to be deleted, when window is closed
      tempFiles.append(lineEdit->text());
   }
}

void kprogramWindow::afterRunningProgram()
{

}

void kprogramWindow::updateProgramProgressForm()
{
   QByteArray programOutput = programProcess.readAll();
   programProgress->ProgramOutputTextEdit->append(QString(programOutput));
   QString timeElapsed;
   timeElapsed.setNum(runTime.elapsed());
   programProgress->RuntimeLabel->setText("Runtime: "+timeElapsed+" ms");
}

void kprogramWindow::processFinished(int exitCode)
{
   programProgress->ProgramOutputTextEdit->append(tr("Finished"));
   QString timeElapsed;
   timeElapsed.setNum(runTime.elapsed());
   programProgress->RuntimeLabel->setText("Runtime: "+timeElapsed+" ms");
   programProgress->CancelButton->setText(tr("Close"));
   if (exitCode==0) {
      if (qobject_cast<kemboss *>(kembossParent)->ksettings.autoCloseProgressWindow())
         programProgress->close();
      afterRunningProgram();

      // Try to display the program output on the 'Output' tab
     if (kpwOutputFileLineEdit != NULL) {
         statusBar()->showMessage(tr("Opening output file"));
         if (kpwOutputTextEdit != NULL && QFile::exists(kpwOutputFileLineEdit->text())) {
            QFile file(kpwOutputFileLineEdit->text());
            if (file.open(QFile::ReadOnly | QFile::Text)) {
               QTextStream in(&file);
               kpwOutputTextEdit->setPlainText(in.readAll());
            }
            kpwOutputTextEdit->document()->setModified(false);
         } else if (kpwOutputGraphicsView != NULL) {
            loadOutputGraphics();
         }
         if (qobject_cast<kemboss *>(kembossParent)->ksettings.autoShowOutputTab() && kpwMainTab != NULL) {
            if (kpwMainTab->tabText(1) == "Output") kpwMainTab->setCurrentIndex(1);
            else if (kpwMainTab->tabText(0) == "Output") kpwMainTab->setCurrentIndex(0);
         }
         statusBar()->clearMessage();
      }
   }
}

void kprogramWindow::loadOutputGraphics()
{
   QString graphicsFile;
   if (QFile::exists(kpwOutputFileLineEdit->text()+".1.png"))
      graphicsFile = kpwOutputFileLineEdit->text()+".1.png";
   else if (QFile::exists(kpwOutputFileLineEdit->text()+".gif"))
      graphicsFile = kpwOutputFileLineEdit->text()+".gif";
   if (!graphicsFile.isEmpty()) {
      if (tempOutputFile == kpwOutputFileLineEdit->text())
         tempFiles.append(graphicsFile);
      QGraphicsScene *scene;
      scene = new QGraphicsScene(this);
      kpwOutputGraphicsView->setScene(scene);
      QPixmap pixmap;
      if (pixmap.load(graphicsFile)) {
         scene->addPixmap(pixmap);
         graphScale = 1;
      } else 
         QMessageBox::warning(this, objectName(), tr("Couldn't load the output graphics!"));
   }
}

void kprogramWindow::killProcess()
{
   if (programProcess.state() != QProcess::NotRunning) programProcess.kill();
   programProgress->close();
}

void kprogramWindow::openSAQForm(const QString num)
{
   SAQForm newSAQForm(this, num);
   newSAQForm.exec();
}

void kprogramWindow::openOAQForm(const QString num)
{
   OAQForm newOAQForm(this, num);
   newOAQForm.exec();
}

void kprogramWindow::openOsAQForm(const QString num)
{
   OsAQForm newOsAQForm(this, num);
   newOsAQForm.exec();
}

void kprogramWindow::openOfAQForm(const QString num)
{
   OfAQForm newOfAQForm(this, num);
   newOfAQForm.exec();
}

void kprogramWindow::openFAQForm(const QString num)
{
   FAQForm newFAQForm(this, num);
   newFAQForm.exec();
}

void kprogramWindow::openGAQForm(const QString num)
{
   GAQForm newGAQForm(this, num);
   newGAQForm.exec();
}

void kprogramWindow::openAAQForm(const QString num)
{
   AAQForm newAAQForm(this, num);
   newAAQForm.exec();
}

void kprogramWindow::addParameter()
{
   ACPForm newACPForm(this);
   newACPForm.exec();
}

void kprogramWindow::copyCommandline()
{
   if (createProgramConfiguration())
      QApplication::clipboard()->setText(qobject_cast<kemboss *>(kembossParent)->ksettings.embossBin()+embossApp()
      #ifdef Q_WS_WIN
      +".exe"
      #endif
      +" "+programConfiguration.join(" "));
}

void kprogramWindow::embossHelp()
{
   assistant(embossApp()+".html");
}

void kprogramWindow::selectOpenFile(const QString &msg, QLineEdit *lineEdit, bool load)
{
   QString fileName = QDir::toNativeSeparators(QFileDialog::getOpenFileName(this, msg, qobject_cast<kemboss *>(kembossParent)->ksettings.currentDir(),"All files (*)"));
   if (!fileName.isEmpty()) {
      lineEdit->setText(fileName);
      qobject_cast<kemboss *>(kembossParent)->ksettings.setcurrentDir(QFileInfo(fileName).absolutePath());
      if (load) loadInputFile(fileName);
   }
}

void kprogramWindow::selectSaveFile(const QString &msg, QLineEdit *lineEdit, bool save)
{
   QString fileName = QDir::toNativeSeparators(QFileDialog::getSaveFileName(this, msg, qobject_cast<kemboss *>(kembossParent)->ksettings.currentDir(),"All files (*)"));
   if (!fileName.isEmpty()) {
      lineEdit->setText(fileName);
      qobject_cast<kemboss *>(kembossParent)->ksettings.setcurrentDir(QFileInfo(fileName).absolutePath());
      if (save) saveInputFile(fileName);
   }
}

QStringList kprogramWindow::loadRebaseEnzymes()
{
   // Load enzyme list from rebase database
   QStringList enzymes;
   QFile file(qobject_cast<kemboss *>(kembossParent)->ksettings.embossData()+"REBASE/embossre.enz");
   if (file.open(QFile::ReadOnly | QFile::Text)) {
      QTextStream in(&file);
      QRegExp sep("\\s+");
      while (!in.atEnd()) {
         QString line = in.readLine();
         if (!(line.startsWith("#") || line.isEmpty()))  // exclude comments and empty lines
            enzymes.append(line.section(sep, 0, 0));
      }
   }
   return enzymes;
}

void kprogramWindow::saveEnzymeList(QListWidget *enzymeList)
{
   QString fileName = QFileDialog::getSaveFileName(this, tr("Save enzyme list"),qobject_cast<kemboss *>(kembossParent)->ksettings.currentDir(),"All files (*)");
   if (fileName.isEmpty()) return;
   qobject_cast<kemboss *>(kembossParent)->ksettings.setcurrentDir(QFileInfo(fileName).absolutePath());

   QFile file(fileName);
   if (!file.open(QFile::WriteOnly | QFile::Text)) {
      QMessageBox::warning(this, objectName(),
                           tr("Cannot save enzyme list %1:\n%2.")
                           .arg(fileName)
                           .arg(file.errorString()));
      return;
   }
   QTextStream out(&file);
   for (int i = 0; i < enzymeList->count(); ++i) {
      out << enzymeList->item(i)->text()+"\n";
   }
}

QString kprogramWindow::selectedEnzymes(QListWidget *enzymeList)
{
   QString enzymes;
   for (int i = 0; i < enzymeList->count(); ++i)
      enzymes += enzymeList->item(i)->text()+",";
   enzymes.chop(1);
   return "-enzymes="+enzymes;
}

QString kprogramWindow::defaultSettingsFile()
{
   return qobject_cast<kemboss *>(kembossParent)->ksettings.kembossDefaultSettings()+"default."+objectName();
}

bool kprogramWindow::maybeSave()
{
   // Check whether the kemboss settings need to be saved
   if (isWindowModified()) {
      QMessageBox::StandardButton ret;
      ret = QMessageBox::warning(this, objectName(),
            tr("The settings have been modified.\n"
            "Do you want to save your changes?"),
            QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
      if (ret == QMessageBox::Save) {
         if (!save()) return false;
      } else if (ret == QMessageBox::Cancel) {
         return false;
      }
   }

   if (!maybeSaveInputFile())
      return false;

   if (!maybeSaveOutputFile())
      return false;

   return true;
}

bool kprogramWindow::maybeSaveInputFile()
// Check whether the input file has been changed
{
   if (kpwInputTextEdit != NULL && kpwInputTextEdit->document()->isModified()) {
      QMessageBox::StandardButton ret;
      ret = QMessageBox::warning(this, objectName(),
            tr("The input data has been modified.\n"
            "Do you want to save your changes?"),
            QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
      if (ret == QMessageBox::Save && kpwInputFileLineEdit != NULL) {
         if (kpwInputFileLineEdit->text().isEmpty()) {
            return on_SaveInputFileAsPushButton_clicked();
         } else {
            return saveInputFile(kpwInputFileLineEdit->text());
         }
      } else if (ret == QMessageBox::Cancel) {
         return false;
      }
   }
   return true;
}

bool kprogramWindow::maybeSaveOutputFile()
// Check whether the output file has been changed or whether there is a temporary output file that has to be saved
{
   if (kpwOutputTextEdit != NULL && kpwOutputTextEdit->document()->isModified()) {
   // The contents of the output text edit as been modified by the user
      QMessageBox::StandardButton ret;
      ret = QMessageBox::warning(this, objectName(),
            tr("The output of the EMBOSS program has been modified.\n"
            "Do you want to save your changes?"),
            QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
      if (ret == QMessageBox::Save && kpwOutputFileLineEdit != NULL) {
         if (kpwOutputFileLineEdit->text().isEmpty() || !tempOutputFile.isEmpty()) {
            return on_SaveOutputFileAsPushButton_clicked();
         } else {
            return saveOutputFile(kpwOutputFileLineEdit->text());
         }
      } else if (ret == QMessageBox::Cancel) {
         return false;
      } else if (ret == QMessageBox::Discard) {
         return true;
      }
   }

   // The output of the EMBOSS program has been saved to a temporary file but has not been changed by the user
   if (!tempOutputFile.isEmpty()) {
      QMessageBox::StandardButton ret;
      ret = QMessageBox::warning(this, objectName(),
            tr("The output of the EMBOSS program has not been saved.\n"
            "Do you want to save the program output?"),
            QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
      if (ret == QMessageBox::Save) {
         return on_SaveOutputFileAsPushButton_clicked();
      } else if (ret == QMessageBox::Cancel) {
         return false;
      }
   }
   return true;
}

void kprogramWindow::deleteTemporaryFiles()
{
   if (!tempInputFile.isEmpty()) {
      if (QFile::exists(tempInputFile))
         QFile::remove(tempInputFile);
      if (kpwInputFileLineEdit != NULL && kpwInputFileLineEdit->text() == tempInputFile)
         kpwInputFileLineEdit->clear();
      tempInputFile.clear();
   }
   if (!tempOutputFile.isEmpty()) {
      if (QFile::exists(tempOutputFile))
         QFile::remove(tempOutputFile);
      if (kpwOutputFileLineEdit != NULL && kpwOutputFileLineEdit->text() == tempOutputFile)
         kpwOutputFileLineEdit->clear();
      tempOutputFile.clear();
   }
   if (!tempFiles.isEmpty()) {
      for (int i = 0; i < tempFiles.size(); ++i) {
         if (QFile::exists(tempFiles.at(i)))
            QFile::remove(tempFiles.at(i));
      }
      tempFiles.clear();
   }
}

void kprogramWindow::closeEvent(QCloseEvent *event)
{
   if (maybeSave()) {
      deleteTemporaryFiles();
      writeSettings();
      qobject_cast<kemboss *>(kembossParent)->programWindows.removeAll(this);
      event->accept();
   } else event->ignore();
}
