/***************************************************************************
 *   Copyright (C) 2007-2010 by Tobias Schwarz                             *
 *   mail@tobias-schwarz.net                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   In addition, as a special exception, the copyright holders give       *
 *   permission to link the code of this program with any edition of       *
 *   the Qt library by Trolltech AS, Norway (or with modified versions     *
 *   of Qt that use the same license as Qt), and distribute linked         *
 *   combinations including the two.  You must obey the GNU General        *
 *   Public License in all respects for all of the code used other than    *
 *   Qt.  If you modify this file, you may extend this exception to        *
 *   your version of the file, but you are not obligated to do so.  If     *
 *   you do not wish to do so, delete this exception statement from        *
 *   your version.                                                         *
 ***************************************************************************/

#include <QtGui>

#include "syntax_highlighter.h"

Highlighter::Highlighter(QTextDocument *parent) : QSyntaxHighlighter(parent)
{
   defineFormats();
   HighlightingRule rule;

   // FASTA sequence name
   rule.pattern = QRegExp("^\\>.*$");
   rule.format = seq_name;
   FASTARules.append(rule);

   // FASTA comment
   rule.pattern = QRegExp("^#.*$");
   rule.format = comment;
   FASTARules.append(rule);

   // FASTA feature
   rule.pattern = QRegExp("/.*$");
   rule.format = comment;
   FASTARules.append(rule);

   // FASTA sequence
   // Everything else is regarded a sequence


   // Alignment title
   rule.pattern = QRegExp("^CLUSTAL.*$");
   rule.format = seq_name;
   AlignmentRules.append(rule);

   // Alignment sequence name
   rule.pattern = QRegExp("^\\>.*$");
   rule.format = seq_name;
   AlignmentRules.append(rule);

   //Alignment score
   rule.pattern = QRegExp("^(Score|Identities| Plus| Minus).*$");
   rule.format = comment;
   AlignmentRules.append(rule);

   // Alignment sequence
   // Everything else is regarded a sequence


   // EMBL sequence name
   rule.pattern = QRegExp("^ID.*$");
   rule.format = seq_name;
   rule.type = 0;
   EMBLRules.append(rule);

   // EMBL sequence description
   rule.pattern = QRegExp("^DE.*$");
   rule.format = seq_name;
   rule.type = 0;
   EMBLRules.append(rule);

   // EMBL + Gb translation start
   AAStart = QRegExp("/translation=\"");

   // EMBL + Gb translation end
   AAEnd = QRegExp("\"");

   // EMBL nucleotide sequence
   rule.pattern = QRegExp("^\\s+[\\w\\s]+\\d+$");
   rule.type = 3;
   EMBLRules.append(rule);


   // Gb header
   rule.pattern = QRegExp("^\\w+");
   rule.format = seq_name;
   rule.type = 0;
   GbRules.append(rule);

   // Gb header 2
   rule.pattern = QRegExp("^\\s{2,8}\\w+");
   rule.format = comment;
   rule.type = 0;
   GbRules.append(rule);

   // Gb sequence
   rule.pattern = QRegExp("^\\s+\\d+\\s[\\w\\s]+$");
   rule.type = 1;
   GbRules.append(rule);


   // AA Sequence rules
   rule.pattern = QRegExp("[cC]+");
   rule.format = cystein;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[AVFPMILavfpmil]+");
   rule.format = apolar_aa;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[STYHNQWstyhnqtw]+");
   rule.format = polar_aa;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[DEde]+");
   rule.format = acidic_aa;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[krKR]");
   rule.format = basic_aa;
   aaSequenceRules.append(rule);

   rule.pattern = QRegExp("[KRkr]{2,}");
   rule.format = dibasic_aa;
   aaSequenceRules.append(rule);


   // Nucleotide sequence rules
   rule.pattern = QRegExp("[aA]+");
   rule.format = a_nuc;
   nucSequenceRules.append(rule);

   rule.pattern = QRegExp("[Tt]+");
   rule.format = t_nuc;
   nucSequenceRules.append(rule);

   rule.pattern = QRegExp("[gG]+");
   rule.format = g_nuc;
   nucSequenceRules.append(rule);

   rule.pattern = QRegExp("[Cc]+");
   rule.format = c_nuc;
   nucSequenceRules.append(rule);
}

void Highlighter::defineFormats()
{
   seq_name.setForeground(Qt::black);
   seq_name.setFontWeight(QFont::Bold);
   comment.setForeground(Qt::gray);
   comment.setFontItalic(true);
   cystein.setFontWeight(QFont::Bold);
   cystein.setForeground(Qt::darkGreen);
   apolar_aa.setForeground(Qt::darkYellow);
   polar_aa.setForeground(Qt::darkGreen);
   acidic_aa.setForeground(Qt::red);
   basic_aa.setForeground(Qt::blue);
   dibasic_aa.setForeground(Qt::blue);
   a_nuc.setForeground(Qt::green);
   t_nuc.setForeground(Qt::red);
   g_nuc.setForeground(Qt::black);
   c_nuc.setForeground(Qt::blue);
}

void Highlighter::setSyntax(const int newSyntax)
{
   syntax = newSyntax;
}

void Highlighter::highlightBlock(const QString &text)
{
   if (syntax == 1) {
   // FASTA AA highlighting
   bool found=0;
   foreach (const HighlightingRule &rule, FASTARules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while (index >= 0) {
         setFormat(index, expression.matchedLength(), rule.format);
         found = 1;
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
   if (found == 0) highlightAASequence(text, 0, text.length());

   } else if (syntax == 2) {
   // FASTA Nuc highlighting
   bool found=0;
   foreach (const HighlightingRule &rule, FASTARules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while (index >= 0) {
         setFormat(index, expression.matchedLength(), rule.format);
         found = 1;
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
   if (found == 0) highlightNucSequence(text, 0, text.length());

   } else if (syntax == 3) {
   // Alignment AA highlighting
   bool found=0;
   foreach (const HighlightingRule &rule, AlignmentRules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while (index >= 0) {
         setFormat(index, expression.matchedLength(), rule.format);
         found = 1;
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
   if (found == 0) highlightAASequence(text, 0, text.length());

   } else if (syntax == 4) {
   // Alignment Nuc highlighting
   bool found=0;
   foreach (const HighlightingRule &rule, AlignmentRules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while (index >= 0) {
         setFormat(index, expression.matchedLength(), rule.format);
         found = 1;
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
   if (found == 0) highlightNucSequence(text, 0, text.length());

   } else if (syntax == 5) {
   // EMBL highlighting
   foreach (const HighlightingRule &rule, EMBLRules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while (index >= 0) {
         if (rule.type == 3) // Sequence is Nuc
            highlightNucSequence(text, index, index + expression.matchedLength());
         else setFormat(index, expression.matchedLength(), rule.format);
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
   setCurrentBlockState(0);
   int startIndex;
   if (previousBlockState() != 1) {
      startIndex = -1;
      startIndex = AAStart.indexIn(text);
      if (startIndex >= 0) startIndex += 14;
   } else startIndex = 2;
   while (startIndex >= 2) {
      int endIndex = AAEnd.indexIn(text, startIndex);
      if (endIndex == -1) {
         setCurrentBlockState(1);
         endIndex = text.length();
      }
      highlightAASequence(text, startIndex, endIndex);
      startIndex = AAStart.indexIn(text, startIndex + 1);
   }

   } else if (syntax == 6) {
   // Genebank AA highlighting
   foreach (const HighlightingRule &rule, GbRules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while (index >= 0) {
         if (rule.type == 1) // Sequence
            highlightAASequence(text, index, index + expression.matchedLength());
         else setFormat(index, expression.matchedLength(), rule.format);
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
   } else if (syntax == 7) {
   // Genebank Nuc highlighting
   foreach (const HighlightingRule &rule, GbRules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while (index >= 0) {
         if (rule.type == 1) // Sequence
            highlightNucSequence(text, index, index + expression.matchedLength());
         else setFormat(index, expression.matchedLength(), rule.format);
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
   setCurrentBlockState(0);
   int startIndex;
   if (previousBlockState() != 1) {
      startIndex = -1;
      startIndex = AAStart.indexIn(text);
      if (startIndex >= 0) startIndex += 14;
   } else startIndex = 2;
   while (startIndex >= 2) {
      int endIndex = AAEnd.indexIn(text, startIndex);
      if (endIndex == -1) {
         setCurrentBlockState(1);
         endIndex = text.length();
      }
      highlightAASequence(text, startIndex, endIndex);
      startIndex = AAStart.indexIn(text, startIndex + 1);
   }
   }
}

void Highlighter::highlightAASequence(const QString &text, const int start, const int end)
{
  foreach (const HighlightingRule &rule, aaSequenceRules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while ((index >= 0) && (index < end)) {
         if (index >=start) setFormat(index, expression.matchedLength(), rule.format);
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
}

void Highlighter::highlightNucSequence(const QString &text, const int start, const int end)
{
   foreach (const HighlightingRule &rule, nucSequenceRules) {
      QRegExp expression(rule.pattern);
      int index = expression.indexIn(text);
      while ((index >= 0) && (index < end)) {
         if (index >=start) setFormat(index, expression.matchedLength(), rule.format);
         index = expression.indexIn(text, index + expression.matchedLength());
      }
   }
}
